/***************************************************************************
 *   Copyright (C) 2007 by Sébastien Laoût                                 *
 *   slaout@linux62.org                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "thememanager.h"

#include "settings.h"
#include "themedialogs.h"

#include <qpixmap.h>
#include <qimage.h>
#include <qdir.h>
#include <qstringlist.h>
#include <kglobal.h>
#include <kglobalsettings.h>
#include <kstandarddirs.h>
#include <ksimpleconfig.h>
#include <kimageeffect.h>

/** Class Theme: */

Theme::Theme(const QString &location, const QString &folderName)
 : m_location(location), m_folderName(folderName), m_backgroundColorOpacity(80), m_backgroundPixmap(0), m_previewPixmap(0)
{
	if (!m_location.endsWith("/"))
		m_location += "/";

	load();
}

Theme::~Theme()
{
	setUsed(false);
}

// Theme informations:
QString Theme::location()                    const { return m_location;  }
QString Theme::folderName()                  const { return m_folderName; }
QString Theme::themeName()                   const { return m_themeName;  }
// Colors:
QColor  Theme::backgroundColor()             const { if (m_backgroundColor.isValid())             return m_backgroundColor;             else return defaultBackgroundColor();             }
int     Theme::backgroundColorOpacity()      const { return m_backgroundColorOpacity; }
QColor  Theme::textColor()                   const { if (m_textColor.isValid())                   return m_textColor;                   else return defaultTextColor();                   }
QColor  Theme::nextPlayingColor()            const { if (m_nextPlayingColor.isValid())            return m_nextPlayingColor;            else return defaultNextPlayingColor();            }
QColor  Theme::dateColor()                   const { if (m_dateColor.isValid())                   return m_dateColor;                   else return defaultDateColor();                   }
QColor  Theme::hourColor()                   const { if (m_hourColor.isValid())                   return m_hourColor;                   else return defaultHourColor();                   }
QColor  Theme::progressBackgroundColor()     const { if (m_progressBackgroundColor.isValid())     return m_progressBackgroundColor;     else return defaultProgressBackgroundColor();     }
QColor  Theme::progressBarColor()            const { if (m_progressBarColor.isValid())            return m_progressBarColor;            else return defaultProgressBarColor();            }
QColor  Theme::progressBackgroundTextColor() const { if (m_progressBackgroundTextColor.isValid()) return m_progressBackgroundTextColor; else return defaultProgressBackgroundTextColor(); }
QColor  Theme::progressBarTextColor()        const { if (m_progressBarTextColor.isValid())        return m_progressBarTextColor;        else return defaultProgressBarTextColor();        }
// Author:
QString Theme::authorName()                  const { return m_authorName;           }
QString Theme::authorEMail()                 const { return m_authorEMail;          }
QString Theme::authorURL()                   const { return m_authorURL;            }
QString Theme::copyrightInformation()        const { return m_copyrightInformation; }
bool    Theme::authorLocked()                const { return m_authorLocked;         }

// Pixmap:
const QPixmap &Theme::backgroundPixmap(int width, int height)
{
	if (m_backgroundPixmap) {
		if (m_backgroundPixmap->width() == width && m_backgroundPixmap->height() == height)
			return *m_backgroundPixmap;

		delete m_backgroundPixmap;
		m_backgroundPixmap = 0;
	}

	QStringList locations;
	if (isWideScreen(width, height))
		locations << "background-wide.png" << "background-wide.jpg" << "background-wide.jpeg" << "background-wide.gif"
		          << "background.png"      << "background.jpg"      << "background.jpeg"      << "background.gif";
	else
		locations << "background.png"      << "background.jpg"      << "background.jpeg"      << "background.gif"
		          << "background-wide.png" << "background-wide.jpg" << "background-wide.jpeg" << "background-wide.gif";
	m_backgroundPixmap = computePixmap(width, height, locations);

	return *m_backgroundPixmap;
}

QPixmap *Theme::computePixmap(int width, int height, const QStringList &locations) const
{
	QPixmap *pixmap = 0;

	if (backgroundColorOpacity() >= 100) {
		pixmap = new QPixmap(width, height);
		pixmap->fill(backgroundColor());
	} else {
		// Load the first adequate image:
		QImage image;
		uint i = 0;
		while (image.isNull() && i < locations.count()) {
			image.load(m_location + locations[i]);
			i++;
		}
		// If no image has been found, do as if background color were 100% opaque:
		if (image.isNull()) {
			pixmap = new QPixmap(width, height);
			pixmap->fill(backgroundColor());
		}
		// If an image has been found, blend:
		else {
			image = image.smoothScale(width, height);
			if (backgroundColorOpacity() > 0)
				KImageEffect::blend(backgroundColor(), image, backgroundColorOpacity() / 100.0);
			pixmap = new QPixmap();
			pixmap->convertFromImage(image);
		}
	}

	return pixmap;
}

const QPixmap &Theme::previewPixmap(int width, int height)
{
	if (m_previewPixmap) {
		if (m_previewPixmap->width() == width && m_previewPixmap->height() == height)
			return *m_previewPixmap;

		delete m_previewPixmap;
		m_previewPixmap = 0;
	}

	double size = 75.0;
	int widthPreview  = (int) (width  >= height ? size : size * width  / height);
	int heightPreview = (int) (height >= width  ? size : size * height / width);

	QStringList locations;
	if (isWideScreen(width, height))
		locations << "background-wide.preview.png" << "background-wide.preview.jpg" << "background-wide.preview.jpeg" << "background-wide.preview.gif"
		          << "background.preview.png"      << "background.preview.jpg"      << "background.preview.jpeg"      << "background.preview.gif";
	else
		locations << "background.preview.png"      << "background.preview.jpg"      << "background.preview.jpeg"      << "background.preview.gif"
		          << "background-wide.preview.png" << "background-wide.preview.jpg" << "background-wide.preview.jpeg" << "background-wide.preview.gif";
	m_previewPixmap = computePixmap(widthPreview, heightPreview, locations);

	return *m_previewPixmap;
}

void Theme::setUsed(bool used) // TODO Rename reset()
{
	if (!used && m_backgroundPixmap || m_previewPixmap) {
		delete m_backgroundPixmap;
		m_backgroundPixmap = 0;
		delete m_previewPixmap;
		m_previewPixmap = 0;
	}
}

void Theme::copyTo(Theme *theme) const
{
	// Theme informations:
	theme->m_location   = m_location;
	theme->m_folderName = m_folderName;
	theme->m_themeName  = m_themeName;
	// Colors:
	theme->m_backgroundColor             = m_backgroundColor;
	theme->m_backgroundColorOpacity      = m_backgroundColorOpacity;
	theme->m_textColor                   = m_textColor;
	theme->m_nextPlayingColor            = m_nextPlayingColor;
	theme->m_dateColor                   = m_dateColor;
	theme->m_hourColor                   = m_hourColor;
	theme->m_progressBackgroundColor     = m_progressBackgroundColor;
	theme->m_progressBarColor            = m_progressBarColor;
	theme->m_progressBackgroundTextColor = m_progressBackgroundTextColor;
	theme->m_progressBarTextColor        = m_progressBarTextColor;
	// Author:
	theme->m_authorName           = m_authorName;
	theme->m_authorEMail          = m_authorEMail;
	theme->m_authorURL            = m_authorURL;
	theme->m_copyrightInformation = m_copyrightInformation;
	theme->m_authorLocked         = m_authorLocked;
//	// Pixmaps:
//	theme->m_backgroundPixmap = m_backgroundPixmap;
//	theme->m_previewPixmap    = m_previewPixmap;

}

// GET:
// Colors:
QColor Theme::backgroundColorSetting()             const { return m_backgroundColor;             }
QColor Theme::textColorSetting()                   const { return m_textColor;                   }
QColor Theme::nextPlayingColorSetting()            const { return m_nextPlayingColor;            }
QColor Theme::dateColorSetting()                   const { return m_dateColor;                   }
QColor Theme::hourColorSetting()                   const { return m_hourColor;                   }
QColor Theme::progressBackgroundColorSetting()     const { return m_progressBackgroundColor;     }
QColor Theme::progressBarColorSetting()            const { return m_progressBarColor;            }
QColor Theme::progressBackgroundTextColorSetting() const { return m_progressBackgroundTextColor; }
QColor Theme::progressBarTextColorSetting()        const { return m_progressBarTextColor;        }

// DEFAULT COLORS:
QColor Theme::defaultBackgroundColor()             const { if (m_textColor.isValid()       &&  tooDark(m_textColor))       return Qt::white; else return Qt::black;   }
QColor Theme::defaultTextColor()                   const { if (m_backgroundColor.isValid() && !tooDark(m_backgroundColor)) return Qt::black; else return Qt::white;   }
QColor Theme::defaultNextPlayingColor()            const { return mixColors(textColor(), backgroundColor()); }
QColor Theme::defaultDateColor()                   const { if (m_hourColor.isValid()) return m_hourColor; else return textColor(); }
QColor Theme::defaultHourColor()                   const { if (m_dateColor.isValid()) return m_dateColor; else return textColor(); }
QColor Theme::defaultProgressBackgroundColor()     const { return KGlobalSettings::baseColor();            }
QColor Theme::defaultProgressBarColor()            const { return KGlobalSettings::highlightColor();       }
QColor Theme::defaultProgressBackgroundTextColor() const { return KGlobalSettings::textColor();            }
QColor Theme::defaultProgressBarTextColor()        const { return KGlobalSettings::highlightedTextColor(); }

// SET:
// Theme informations:
void Theme::setThemeName(const QString &name)                   { m_themeName                   = name;        }
// Colors:
void Theme::setBackgroundColor(const QColor &color)             { m_backgroundColor             = color;       }
void Theme::setBackgroundColorOpacity(int opacity)              { m_backgroundColorOpacity      = opacity;     }
void Theme::setTextColor(const QColor &color)                   { m_textColor                   = color;       }
void Theme::setNextPlayingColor(const QColor &color)            { m_nextPlayingColor            = color;       }
void Theme::setDateColor(const QColor &color)                   { m_dateColor                   = color;       }
void Theme::setHourColor(const QColor &color)                   { m_hourColor                   = color;       }
void Theme::setProgressBackgroundColor(const QColor &color)     { m_progressBackgroundColor     = color;       }
void Theme::setProgressBarColor(const QColor &color)            { m_progressBarColor            = color;       }
void Theme::setProgressBackgroundTextColor(const QColor &color) { m_progressBackgroundTextColor = color;       }
void Theme::setProgressBarTextColor(const QColor &color)        { m_progressBarTextColor        = color;       }
// Author:
void Theme::setAuthorName(const QString &name)                  { m_authorName                  = name;        }
void Theme::setAuthorEMail(const QString &eMail)                { m_authorEMail                 = eMail;       }
void Theme::setAuthorURL(const QString &url)                    { m_authorURL                   = url;         }
void Theme::setCopyrightInformation(const QString &information) { m_copyrightInformation        = information; }
void Theme::setAuthorLocked(bool lock)                          { m_authorLocked                = lock;        }

void Theme::load()
{
	KSimpleConfig config(m_location + "kirocker-theme.config", /*readOnly=*/true);

	QColor *defaultColor = new QColor();

	config.setGroup("Theme");
	m_themeName = config.readEntry("Name", "");

	config.setGroup("Colors");
	m_backgroundColor             = config.readColorEntry("Background",             defaultColor);
	m_backgroundColorOpacity      = config.readNumEntry(  "BackgroundOpacity",      80);
	m_textColor                   = config.readColorEntry("Text",                   defaultColor);
	m_nextPlayingColor            = config.readColorEntry("NextPlaying",            defaultColor);
	m_dateColor                   = config.readColorEntry("Date",                   defaultColor);
	m_hourColor                   = config.readColorEntry("Hour",                   defaultColor);
	m_progressBackgroundColor     = config.readColorEntry("ProgressBackground",     defaultColor);
	m_progressBarColor            = config.readColorEntry("ProgressBar",            defaultColor);
	m_progressBackgroundTextColor = config.readColorEntry("ProgressBackgroundText", defaultColor);
	m_progressBarTextColor        = config.readColorEntry("ProgressBarText",        defaultColor);

	config.setGroup("Author");
	m_authorName           = config.readEntry("Name",                 "");
	m_authorEMail          = config.readEntry("EMail",                "");
	m_authorURL            = config.readEntry("URL",                  "");
	m_copyrightInformation = config.readEntry("CopyrightInformation", "");
	m_authorLocked         = config.readBoolEntry("Locked",           false);

	delete defaultColor;
	defaultColor = 0;
}

void Theme::save()
{
	KSimpleConfig config(m_location + "kirocker-theme.config", /*readOnly=*/false);

	config.setGroup("Theme");
	config.writeEntry("Name",                   m_themeName);

	config.setGroup("Colors");
	config.writeEntry("Background",             m_backgroundColor);
	config.writeEntry("BackgroundOpacity",      m_backgroundColorOpacity);
	config.writeEntry("Text",                   m_textColor);
	config.writeEntry("NextPlaying",            m_nextPlayingColor);
	config.writeEntry("Date",                   m_dateColor);
	config.writeEntry("Hour",                   m_hourColor);
	config.writeEntry("ProgressBackground",     m_progressBackgroundColor);
	config.writeEntry("ProgressBar",            m_progressBarColor);
	config.writeEntry("ProgressBackgroundText", m_progressBackgroundTextColor);
	config.writeEntry("ProgressBarText",        m_progressBarTextColor);

	config.setGroup("Author");
	config.writeEntry("Name",                   m_authorName);
	config.writeEntry("EMail",                  m_authorEMail);
	config.writeEntry("URL",                    m_authorURL);
	config.writeEntry("CopyrightInformation",   m_copyrightInformation);
	config.writeEntry("Locked",                 m_authorLocked);

	setUsed(false);
}

bool Theme::tooDark(const QColor &color)
{
	int dontCare, value;
	color.getHsv(/*hue:*/dontCare, /*saturation:*/dontCare, value);
	return value < 175;
}

QColor Theme::mixColors(const QColor &color1, const QColor &color2)
{
	QColor mixedColor;
	mixedColor.setRgb( (color1.red()   + color2.red())   / 2,
	                   (color1.green() + color2.green()) / 2,
	                   (color1.blue()  + color2.blue())  / 2 );
	return mixedColor;
}

bool Theme::isWideScreen(int width, int height)
{
	double ratio = height / (double) width;
	return !(ratio > 0.7 && ratio < 0.8); // Not wide screen if ratio is roughly 3/4 (== 0.75)
}

Theme *Theme::current()
{
	return ThemeManager::instance()->currentTheme();
}

/*bool operator<(const Theme &theme1, const Theme &theme2)
{
	return theme1.themeName().compare(theme2.themeName());
}*/

/** Class ThemeList: */

ThemeList::ThemeList()
 : QPtrList<Theme>()
{
}

int ThemeList::compareItems(QPtrCollection::Item item1, QPtrCollection::Item item2)
{
	Theme *theme1 = (Theme *) item1;
	Theme *theme2 = (Theme *) item2;
	return theme1->themeName().lower().compare(theme2->themeName().lower());
}


/** Class ThemeManager: */

ThemeManager *ThemeManager::s_instance = 0;

ThemeManager::ThemeManager()
 : QObject(), m_currentTheme(0)
{
}

ThemeManager::~ThemeManager()
{
}

void ThemeManager::loadThemeList()
{
	QString currentThemeFolderName = Settings::currentTheme();

	// Load Theme List:
	QStringList directories = KGlobal::dirs()->resourceDirs("data"); // eg. { "/home/seb/.kde/share/apps/", "/usr/share/apps/" }
	// For each data folder:
	for (QStringList::Iterator it = directories.begin(); it != directories.end(); ++it) {
		// For each theme folder in this data folder:
		QDir dir(*it + "kirocker/themes/", /*nameFilder=*/"", /*sortSpec=*/QDir::Name | QDir::IgnoreCase, /*filterSpec=*/QDir::Dirs | QDir::NoSymLinks);
		QStringList files = dir.entryList();
		for (QStringList::Iterator it2 = files.begin(); it2 != files.end(); ++it2)
			if (*it2 != "." && *it2 != "..") {
				Theme *theme = new Theme(*it + "kirocker/themes/" + *it2, *it2);
				m_themes.inSort(theme);
				if (*it2 == currentThemeFolderName)
					m_currentTheme = theme;
			}
	}

	// Load the Current Theme (from config file, default one, or first in the list):
	if (!m_currentTheme && m_themes.count() > 0)
		m_currentTheme = m_themes.at(0); // TODO

	// Load a Default Theme if none in the list:
	if (!m_currentTheme)
		m_currentTheme = new Theme("", "");
}

Theme *ThemeManager::currentTheme()
{
	if (!m_currentTheme)
		loadThemeList();

	return m_currentTheme;
}

ThemeManager *ThemeManager::instance()
{
	if (!s_instance)
		s_instance = new ThemeManager();

	return s_instance;
}

const ThemeList &ThemeManager::themes() const
{
	return m_themes;
}

void ThemeManager::setTheme(Theme *theme)
{
	m_currentTheme = theme;
	emit themeChanged();
	Settings::setCurrentTheme(theme->folderName());
	Settings::writeConfig();
}

#include <kmessagebox.h>

void ThemeManager::chooseTheme(int screenWidth, int screenHeight, QWidget *parent)
{
	static ThemeChooserDialog *dialog = 0;
	static int lastScreenWidth  = 0;
	static int lastScreenHeight = 0;

	if (dialog == 0 || lastScreenWidth != screenWidth || lastScreenHeight != screenHeight) {
		if (dialog != 0) {
			delete dialog;
			dialog = 0;
		}

		KMessageBox::information(parent,
			"<p>This is a <b>Beta Version</b>. It is not finished and contains some glitches.</p>"
			"<p>Please give your impression on <a href=\"http://www.kde-apps.org/content/show.php/Kirocker+Music+Display?content=52869\">Kirocker Music Display page</a> "
			"about the themes: which one(s) do you like, which one(s) should be removed or moved to an external package...</p>"
			"<p>The graphical editor is not finished yet. To add or modify themes, please visit <a href=\"http://slaout.linux62.org/kirocker/theming.html\">the tutorial page</a>. "
			"You're also welcome to post your creations on the first link. If they are good they may be included by default in the final version.</p>",
			"Beta Version", "dontShowAgain40Beta", KMessageBox::Notify | KMessageBox::AllowLink);

		dialog = new ThemeChooserDialog(screenWidth, screenHeight, parent);
		lastScreenWidth  = screenWidth;
		lastScreenHeight = screenHeight;
	}

	dialog->exec();

	// FIXME This recenters the dialog, and select the current theme:
	delete dialog;
	dialog = 0;
}

#include "thememanager.moc"
